#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <time.h>
#include <pthread.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include "globals.h"
#include "synchronize.h"

// Buffer mutex
Mutex buffer_mutex;

// Semaphores
Semaphore empty_slots_available;
Semaphore filled_slots_available;

void consume();

// Critical region.
void remove_item(Item *item);

// Pointer to shared memory object.
Shared *shared;

int main(int argc, char *argv[])
{
    if (argc != 2)
    {
        fprintf(stderr, "Usage: id\n");
        return -1;
    }
    int id = atoi(argv[1]);

    int shm_fd;      // shared memory file descriptor

    // Open the shared memory object and map it.
    shm_fd = shm_open(SHARED_MEMORY_NAME, O_CREAT | O_RDWR, 0666);
    shared = mmap(0, SHARED_SIZE, PROT_WRITE, MAP_SHARED, shm_fd, 0);

    // Open existing semaphores.
    printf(" Empty slots available semaphore opened: %p\n",
           (char *)OPEN(&empty_slots_available, EMPTY_SEMAPHORE_NAME));
    printf("Filled slots available semaphore opened: %p\n",
           (char *)OPEN(&filled_slots_available, FILLED_SEMAPHORE_NAME));

    // Use the existing buffer mutex lock.
    printf("                      Buffer mutex used: %p\n",
           (char *)USE(&buffer_mutex, "buffer_mutex"));
    printf("\n");

    consume(id);

    printf("\nDone!\n");
    return 0;
}

void consume(int id)
{
    srand(time(0));
    Item item;

    for (;;)
    {
        sleep(rand()%2);

        remove_item(&item);

        printf("Consumer %d removed %5d\n", id, item);
        fflush(stdout);
    }
}

void remove_item(Item *item)
{
    WAIT(&filled_slots_available);

    // Critical region: Remove an item from the buffer.
    LOCK(&buffer_mutex);
        *item = shared->buffer[shared->out];
        shared->out = (shared->out + 1)%BUFFER_SIZE;
    UNLOCK(&buffer_mutex);

    SIGNAL(&empty_slots_available);
}
